#!/usr/bin/env python

# multi_timed_run.py by jesse ruderman
# based on timed_run.py by coop and bob clary

# Usage: af_timed_run timeout firefoxpath

# For example:
# ./af_timed_run.py 240 "/Users/admin/Desktop/builds/Minefield 2006-09-24.app/Contents/MacOS/firefox-bin"
# ./af_timed_run.py 240 "/Users/admin/trunk/mozilla/fx-debug-shared-gcc401/dist/MinefieldDebug.app/Contents/MacOS/firefox-bin"

# How to interpret various outputs:
#   NORMAL *with output* -- usually indicates that the fuzzer made the javascript engine halt execution (out of memory?)
#   CRASHED              -- on Mac OS X, use Console.app to view ~/Library/Logs/CrashReporter/js.log to see the stack
#   TIMED OUT            -- hang or very slow loop
#   ABNORMAL             -- a fuzzer-generated function did something that let to a TypeError in the fuzzer itself


exitOSError   = 66
exitInterrupt = 99

(CRASHED, TIMED_OUT, NORMAL, ABNORMAL, NONE) = range(5)


import os, signal, sys, time, random

pid = None
URLs = []


def alarm_handler(signum, frame):
    global pid
    # print "TIMED OUT!"
    try:
        os.kill(pid, signal.SIGKILL)
    except:
        pass

def forkexec(command, args, stdoutfn):
    try:
        newpid = os.fork()
        if newpid == 0:  # Child

            # redirect stdout (just for the child!).
            # i hope i'm not supposed to close() this file somehow...
            so = file(stdoutfn, 'w')
            os.dup2(so.fileno(), sys.stdout.fileno())

            # redirect stderr too (to a separate file, for now)
            so = file("err-" + stdoutfn, 'w')
            os.dup2(so.fileno(), sys.stderr.fileno())
            
            # transfer control of the child to the target program
            os.execvp(command, args)
             
        else:  # Parent
            return newpid
    except OSError, e:
        print "ERROR: %s %s failed: %d (%s)" % (command, args, e.errno, e.strerror)
        sys.exit(exitOSError)

def one_timed_run(logfilename, URL):
    global pid
    global CRASHED, TIMED_OUT, NORMAL, ABNORMAL, NONE

    signal.signal(signal.SIGALRM, alarm_handler)
    signal.alarm(int(sys.argv[1]))

    starttime = time.time()
    
    prog = sys.argv[2]
    args = [prog, URL]

    msg = ''

    try:
      pid = forkexec(prog, args, logfilename)

      try: 
        status = os.waitpid(pid, 0)[1]
      except OSError:
        msg = 'TIMED OUT' # assume this means it timed out
        sta = TIMED_OUT
      
      signal.alarm(0) # Cancel the alarm

    except KeyboardInterrupt:
      print "Bye!"
      try:
        os.kill(pid, signal.SIGKILL)
      except OSError, e:
        print "Unable to kill it: %s" % e
      sys.exit(exitInterrupt)

    stoptime = time.time()
    elapsedtime = stoptime - starttime
    
    if (msg == 'TIMED OUT'):
        pass
    elif os.WIFEXITED(status):
        rc = os.WEXITSTATUS(status)
        if rc == 0:
            msg = 'NORMAL'
            sta = NORMAL
        else:
            # the original timed_run.py considered rc >= 3 to be a crash. why?
            msg = 'ABNORMAL ' + str(rc)
            sta = ABNORMAL
    elif os.WIFSIGNALED(status):
        print time.asctime()
        msg = 'CRASHED signal %d' % os.WTERMSIG(status)
        sta = CRASHED
    else:
        msg = 'NONE'
        sta = NONE

    print "EXIT STATUS: %s (%f seconds) (log file %s)" % (msg, elapsedtime, logfilename)

    return sta

def showImportantBits(logfilename):
    cmd = "grep fuzzer- %s" % logfilename
    os.system(cmd)

def many_timed_runs():
    global CRASHED, TIMED_OUT, NORMAL, ABNORMAL, NONE
    global URLs
    
    iteration = 0
    while True:
        iteration += 1
        URL = random.choice(URLs) + "#squarefree-af"
        logfilename = "w%d" % iteration
        sta = one_timed_run(logfilename, URL)
        # if not (sta == NORMAL):
        if ((sta == CRASHED) or (sta == ABNORMAL)):
             print URL
             showImportantBits(logfilename)
             
def getURLs():
    global URLs
    
    urlfile = open("urls", "r")
    for line in urlfile:
        if (not line.startswith("#") and len(line) > 2):
            URLs.append(line.rstrip())

getURLs()

many_timed_runs()